package org.pdf.filigrane;

import com.drew.imaging.ImageMetadataReader;
import com.drew.metadata.exif.ExifIFD0Directory;
import com.twelvemonkeys.image.ImageUtil;
import org.apache.pdfbox.Loader;
import org.apache.pdfbox.io.IOUtils;
import org.apache.pdfbox.pdmodel.PDDocument;
import org.apache.pdfbox.pdmodel.PDPage;
import org.apache.pdfbox.pdmodel.PDPageContentStream;
import org.apache.pdfbox.pdmodel.PDPageTree;
import org.apache.pdfbox.pdmodel.common.PDRectangle;
import org.apache.pdfbox.pdmodel.graphics.image.PDImageXObject;
import org.apache.pdfbox.rendering.ImageType;
import org.apache.pdfbox.rendering.PDFRenderer;
import org.apache.pdfbox.tools.imageio.ImageIOUtil;

import javax.imageio.ImageIO;
import java.awt.*;
import java.awt.geom.AffineTransform;
import java.awt.image.BufferedImage;
import java.awt.image.ConvolveOp;
import java.awt.image.Kernel;
import java.io.*;
import java.util.List;
import java.util.*;
import java.util.concurrent.ThreadLocalRandom;


public class Main {
    private PdfTemplateParameters params = PdfTemplateParameters.builder().build();

    private final boolean shouldUseDistortion = false;
    private final boolean shouldUseColors = false;

    public static void main(String[] args) throws Exception {
        if (args.length < 3) {
            System.out.println("[dest] [watermak] [input1] ...");
            return;
        }

        var dest = args[0];
        var watermark = args[1];
        var input = new ArrayList<EntryFile>();

        for (int i = 2; i < args.length; i++) {
            var path = args[i];
            input.add(new EntryFile(new FileInputStream(path), path.endsWith(".pdf")));
        }

        var bytes = new Main().render(input, watermark);

        var destFile = new FileOutputStream(dest);
        destFile.write(bytes);
        destFile.flush();
        destFile.close();
    }

    public byte[] render(List<EntryFile> data, String watermarkText) throws IOException {
        try (PDDocument document = new PDDocument()) {

            data.stream()
                    .map(this::convertToImages)
                    .flatMap(Collection::stream)
                    .map(this::smartCrop)
                    .filter(Objects::nonNull)
                    .map(this::fitImageToPage)
                    .map(bim -> applyWatermark(bim, watermarkText))
                    .forEach(bim -> addImageAsPageToDocument(document, bim));

            try (ByteArrayOutputStream baos = new ByteArrayOutputStream()) {
                document.save(baos);
                return baos.toByteArray();
            }
        } catch (IOException e) {
            System.err.println("Exception while generate BO PDF documents" + e);
            throw e;
        }
    }

    private final Color[] COLORS = {
            new Color(64, 64, 64, 255),
            new Color(32, 32, 32, 220),
            new Color(0, 0, 0, 110),
            new Color(0, 0, 91, 170),
            new Color(255, 0, 0, 170)
    };


    /**
     * Convert PDF to image - let other type unchanged
     *
     * @param file source
     * @return list of extracted images
     */
    private List<BufferedImage> convertToImages(EntryFile file) {
        try {
            if (file.isPDF) {
                List<BufferedImage> images = new ArrayList<>();
                try (PDDocument document = Loader.loadPDF(file.is.readAllBytes())) {
                    PDFRenderer pdfRenderer = new PDFRenderer(document);
                    PDPageTree pagesTree = document.getPages();
                    for (int i = 0; i < pagesTree.getCount(); i++) {
                        PDRectangle pageMediaBox = pagesTree.get(i).getMediaBox();
                        float ratioImage = pageMediaBox.getHeight() / pageMediaBox.getWidth();
                        float ratioPDF = params.mediaBox.getHeight() / params.mediaBox.getWidth();

                        // scale according the greater axis
                        PageDimension dimension = (ratioImage < ratioPDF) ?
                                new PageDimension((int) pageMediaBox.getWidth(), (int) (pageMediaBox.getWidth() * ratioPDF), 0)
                                : new PageDimension((int) (pageMediaBox.getHeight() / ratioPDF), (int) pageMediaBox.getHeight(), 0);


                        float scale = (dimension.width < params.maxPage.width) ? 1f :
                                params.maxPage.width / pageMediaBox.getWidth();// image is too big - scale if necessary

                        // x2 - double the image resolution (prevent quality loss if image is cropped)
                        images.add(pdfRenderer.renderImage(i, scale * 2, ImageType.RGB));
                    }
                    return images;
                } catch (Exception e) {
                    System.err.println("Exception while converting pdf page to image" + e);
                    return images;
                }
            }

            return Collections.singletonList(createImageWithOrientation(file));

        } catch (IOException e) {
            throw new RuntimeException("Unable to convert pdf to image", e);
        }
    }

    private BufferedImage createImageWithOrientation(EntryFile file) throws IOException {
        // duplicate input stream - because it will be read twice
        var baos = new ByteArrayOutputStream();
        IOUtils.copy(file.is, baos);
        byte[] bytes = baos.toByteArray();

        try (var imageInput = new ByteArrayInputStream(bytes)) {
            BufferedImage image = ImageIO.read(imageInput);

            // check en rotate according metadata
            try (ByteArrayInputStream imageInputForMeta = new ByteArrayInputStream(bytes)) {

                var metadata = ImageMetadataReader.readMetadata(imageInputForMeta);
                var dir = metadata.getFirstDirectoryOfType(ExifIFD0Directory.class);
                int orientation = dir != null && dir.getInteger(ExifIFD0Directory.TAG_ORIENTATION) != null
                        ? dir.getInt(ExifIFD0Directory.TAG_ORIENTATION) : 0;

                return switch (orientation) {
                    case 2 -> ImageUtil.createFlipped(image, ImageUtil.FLIP_HORIZONTAL);
                    case 3 -> ImageUtil.createRotated(image, ImageUtil.ROTATE_180);
                    case 4 -> ImageUtil.createFlipped(image, ImageUtil.FLIP_VERTICAL);
                    case 5 -> ImageUtil.createRotated(
                            ImageUtil.createFlipped(image, ImageUtil.FLIP_VERTICAL),
                            ImageUtil.ROTATE_90_CW);
                    case 6 -> ImageUtil.createRotated(image, ImageUtil.ROTATE_90_CW);
                    case 7 -> ImageUtil.createRotated(
                            ImageUtil.createFlipped(image, ImageUtil.FLIP_VERTICAL),
                            ImageUtil.ROTATE_90_CCW);
                    case 8 -> ImageUtil.createRotated(image, ImageUtil.ROTATE_90_CCW);
                    default -> image; // 0,1 included
                };

            } catch (Exception e) {
                System.err.println("Unable to rotate and flip from metadata" + e);
            }
            return image;
        }
    }

    /**
     * Apply document crop if require
     */
    protected BufferedImage smartCrop(BufferedImage image) {
        // by default there is not crop on BO Documents
        return image;
    }

    /**
     * Apply watermark - source image should already have good ratio
     *
     * @param bim source image
     * @return result image
     */
    public BufferedImage applyWatermark(BufferedImage bim, String watermarkText) {
        try {
            //Create a watermark layer
            int diagonal = (int) Math.sqrt(bim.getWidth() * bim.getWidth() + bim.getHeight() * bim.getHeight());

            BufferedImage watermarkLayer = new BufferedImage(diagonal, diagonal, BufferedImage.TYPE_INT_ARGB);

            Graphics2D g = watermarkLayer.createGraphics();

            String watermark = watermarkText.repeat(1 + (128 / watermarkText.length()));

            g.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, ThreadLocalRandom.current().nextFloat(0.52f, 0.6f)));
            g.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);


            // allows to have small variation on the watermark position at each generation
            float spaceBetweenText = diagonal / ThreadLocalRandom.current().nextFloat(8f, 10f);
            for (int i = 1; i < 11; i++) {
                Font font = new Font("Arial", Font.PLAIN, 28 * bim.getWidth() / params.maxPage.width);
                if (shouldUseColors) {
                    g.setColor(COLORS[ThreadLocalRandom.current().nextInt(0, COLORS.length)]);
                } else {
                    g.setColor(Color.DARK_GRAY);
                }
                g.setFont(font);
                g.drawString(watermark, 0, i * spaceBetweenText);
            }

            // Create a gaussian blur layer
            int radius = ThreadLocalRandom.current().nextInt(45, 65);

            BufferedImage blurredTextLayer = new BufferedImage(diagonal, diagonal, BufferedImage.TYPE_INT_ARGB);
            Graphics2D blurredTextLayerGraphics = blurredTextLayer.createGraphics();
            blurredTextLayerGraphics.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, ThreadLocalRandom.current().nextFloat(0.75f, 0.95f)));
            blurredTextLayerGraphics.drawImage(watermarkLayer, 0, 0, null);
            blurredTextLayer = getGaussianBlurFilter(radius, true).filter(blurredTextLayer, null);
            blurredTextLayer = getGaussianBlurFilter(radius, false).filter(blurredTextLayer, null);
            blurredTextLayerGraphics.dispose();

            // Merge layers
            Graphics2D gf = bim.createGraphics();
            gf.drawImage(bim, 0, 0, null);

            BufferedImage buffer;
            if (shouldUseDistortion) {
                DFFilter filter = new DFFilter();
                buffer = filter.filter(watermarkLayer, null);
            } else {
                buffer = watermarkLayer;
            }

            BufferedImage rotated = new BufferedImage(diagonal, diagonal, buffer.getType());
            Graphics2D graphic = rotated.createGraphics();
            graphic.rotate(Math.toRadians(-25), diagonal / 2f, diagonal / 2f);
            graphic.drawImage(buffer, null, 0, 0);
            graphic.drawImage(blurredTextLayer, 0, 0, null);
            graphic.dispose();

            BufferedImage cropedRotated = rotated.getSubimage(diagonal / 2 - bim.getWidth() / 2, diagonal / 2 - bim.getHeight() / 2, diagonal / 2 + bim.getWidth() / 2, diagonal / 2 + bim.getHeight() / 2);

            gf.drawImage(cropedRotated, 0, 0, null);

            gf.dispose();

            return bim;
        } catch (Exception e) {
            e.printStackTrace();
            throw new RuntimeException("Unable to fit image to the page", e);
        }
    }

    /**
     * Fit image to page dimension - A4 with 128 DPI
     *
     * @param bim
     * @return image fit to the page
     */
    private BufferedImage fitImageToPage(BufferedImage bim) {
        try {
            float ratioImage = bim.getHeight() / (float) bim.getWidth();
            float ratioPDF = params.mediaBox.getHeight() / params.mediaBox.getWidth();

            // scale according the greater axis
            PageDimension dimension = (ratioImage < ratioPDF) ?
                    new PageDimension(bim.getWidth(), (int) (bim.getWidth() * ratioPDF), 0)
                    : new PageDimension((int) (bim.getHeight() / ratioPDF), bim.getHeight(), 0);


            float scale = (dimension.width < params.maxPage.width) ? 1f :
                    params.maxPage.width / (float) bim.getWidth();// image is too big - scale if necessary

            // translate in center and scale if necessary
            var affineTransform = new AffineTransform();
            affineTransform.translate(scale * (dimension.width - bim.getWidth()) / 2, scale * (dimension.height - bim.getHeight()) / 2);
            affineTransform.scale(scale, scale);

            // Draw the image on to the buffered image
            BufferedImage resultImage = new BufferedImage((int) (scale * dimension.width), (int) (scale * dimension.height), BufferedImage.TYPE_INT_RGB);

            Graphics2D g = resultImage.createGraphics();
            g.setColor(Color.WHITE);
            g.fillRect(0, 0, (int) (scale * dimension.width), (int) (scale * dimension.height));
            g.drawImage(bim, affineTransform, null);
            g.dispose();

            return resultImage;
        } catch (Exception e) {
            e.printStackTrace();
            throw new RuntimeException("Unable to fit image to the page", e);
        }
    }

    /**
     * Add A4 page to Document from image.
     *
     * @param document document destination
     * @param bim      image to include
     */
    private void addImageAsPageToDocument(PDDocument document, BufferedImage bim) {
        var page = new PDPage(PDRectangle.A4);
        document.addPage(page);

        try (ByteArrayOutputStream out = new ByteArrayOutputStream()) {
            ImageIOUtil.writeImage(bim, "jpg", out, params.maxPage.dpi, params.compressionQuality);


            var pdImage = PDImageXObject.createFromByteArray(document, out.toByteArray(), "");
            try (PDPageContentStream contentStream = new PDPageContentStream(document, page, PDPageContentStream.AppendMode.OVERWRITE, true, true)) {
                contentStream.drawImage(pdImage, 0, 0, PDRectangle.A4.getWidth(), bim.getHeight() * PDRectangle.A4.getWidth() / bim.getWidth());
            }
        } catch (IOException e) {
            e.printStackTrace();
            throw new RuntimeException("Unable to write image");
        }
    }

    private static ConvolveOp getGaussianBlurFilter(int radius, boolean horizontal) {
        int size = radius * 2 + 1;
        float[] data = new float[size];

        float sigma = radius / 3.0f;
        float twoSigmaSquare = 2.0f * sigma * sigma;
        float sigmaRoot = (float) Math.sqrt(twoSigmaSquare * Math.PI);
        float total = 0.0f;

        for (int i = -radius; i <= radius; i++) {
            float distance = i * i;
            int index = i + radius;
            data[index] = (float) Math.exp(-distance / twoSigmaSquare) / sigmaRoot;
            total += data[index];
        }
        for (int i = 0; i < data.length; i++) {
            data[i] /= total;
        }
        Kernel kernel = (horizontal) ? new Kernel(size, 1, data) : new Kernel(1, size, data);

        return new ConvolveOp(kernel, ConvolveOp.EDGE_NO_OP, null);
    }
}